/*
This class extends asynMotorAxis by some features used in SINQ.

Stefan Mathis, November 2024
*/

#ifndef __SINQDRIVER
#define __SINQDRIVER
#include "asynMotorAxis.h"

class epicsShareClass sinqAxis : public asynMotorAxis {
  public:
    /**
     * @brief Construct a new sinqAxis object
     *
     * @param pC_                 Pointer to the controller of the axis
     * @param axis                Index of the axis
     */
    sinqAxis(class sinqController *pC_, int axisNo);

    /**
     * @brief This function is executed once during the very first poll.
     *
     * This function is executed at the very first poll after the IOC startup.
    If it returns anything else than 'asynSuccess', the function is evaluated
    again after 100 ms until it succeeds. Every 10 trials a warning is emitted.
    The default implementation just returns asynSuccess and is meant to be
    overwritten by concrete driver implementations.
     *
     * @return asynStatus
     */
    virtual asynStatus atFirstPoll();

    /**
     * @brief Perform some standardized operation before and after the concrete
    `doPoll` implementation.
     *
     * Wrapper around doPoll which performs the following operations:
    Before calling doPoll:
    - Try to execute atFirstPoll once (and retry, if that failed)

    After calling doPoll:
    - Reset motorStatusProblem_, motorStatusCommsError_ and motorMessageText_ if
    doPoll returned asynSuccess
    - If the movement timeout watchdog has been started, check it.
    - Update the parameter library entry motorEnableRBV_ by calling isEnabled.
    - Run `callParamCallbacks`
    - Return the status of `doPoll`
     *
     * @param moving              Forwarded to `doPoll`.
     * @return asynStatus         Forward the status of `doPoll`, unless one of
    the parameter library operation fails (in that case, returns the failed
    operation status).
     */
    asynStatus poll(bool *moving);

    /**
     * @brief Implementation of the "proper", device-specific poll method. This
    method should be implemented by a child class of sinqAxis.
     *
     * @param moving            Should be set to true, if the axis is moving,
     * and false otherwise.
     * @return asynStatus
     */
    virtual asynStatus doPoll(bool *moving);

    /**
     * @brief Perform some standardized operation before and after the concrete
    `doMove` implementation.

    * Wrapper around `doMove` which calculates the (absolute) target position
    and stores it in the parameter library. After that, it calls and returns
    `doMove`.
     *
     * @param position            Forwarded to `doMove`.
     * @param relative            Forwarded to `doMove`.
     * @param minVelocity         Forwarded to `doMove`.
     * @param maxVelocity         Forwarded to `doMove`.
     * @param acceleration        Forwarded to `doMove`.
     * @return asynStatus         Forward the status of `doMove`, unless one of
    the parameter library operation fails (in that case, returns the failed
    operation status).
     */
    asynStatus move(double position, int relative, double minVelocity,
                    double maxVelocity, double acceleration);

    /**
     * @brief Implementation of the "proper", device-specific move method. This
    method should be implemented by a child class of sinqAxis.
     *
     * @param position            Target position `VAL` from the motor record
     * @param relative            Specifies, whether the target position is
    relative or absolute.
     * @param minVelocity         Minimum velocity VMIN from the motor record
     * @param maxVelocity         Maximum velocity VMAX from the motor record
     * @param acceleration        Acceleration ACCEL from the motor record
     * @return asynStatus
     */
    virtual asynStatus doMove(double position, int relative, double minVelocity,
                              double maxVelocity, double acceleration);

    /**
     * @brief Perform some standardized operation before and after the concrete
    `doHome` implementation.
    *
    * Wrapper around move which calculates the (absolute) target position and
    stores it in the parameter library. The target position in a homing maneuver
    is calculated as follows:

    if abs(current position - high limit) > abs(current position - low limit)
    {
    high limit
    }
    else
    {
    low limit
    }

    After that, it calls and returns doHome.
     *
     * @param minVelocity         Forwarded to `doHome`.
     * @param maxVelocity         Forwarded to `doHome`.
     * @param acceleration        Forwarded to `doHome`.
     * @param forwards            Forwarded to `doHome`.
     * @return asynStatus         Forward the status of `doHome`, unless one of
    the parameter library operation fails (in that case, returns the failed
    operation status).
     */
    asynStatus home(double minVelocity, double maxVelocity, double acceleration,
                    int forwards);

    /**
     * @brief Implementation of the "proper", device-specific home method. This
    method should be implemented by a child class of sinqAxis.
     *
     * @param minVelocity         Minimum velocity VMIN from the motor record
     * @param maxVelocity         Maximum velocity VMAX from the motor record
     * @param acceleration        Acceleration ACCEL from the motor record
     * @param forwards            Is 1, if the motor record field HOMF was used
    to trigger the homing, and 0, if HOMR was used.
     * @return asynStatus
     */
    virtual asynStatus doHome(double minVelocity, double maxVelocity,
                              double acceleration, int forwards);

    /**
     * @brief This function enables / disables an axis. It should be implemented
     * by a child class of sinqAxis.
     *
     * @param on
     * @return asynStatus
     */
    virtual asynStatus enable(bool on);

    /**
     * @brief Populate the motor record fields VELO, VBAS and VMAX
     *
     * Populates the speed fields of the motor record. If the param lib
     * entry motorCanSetSpeed_ (connected to the PV x:VariableSpeed) is set to
     * 1, VBAS and VMAX are set to min and max respectively. Otherwise, they are
     * set to val. Additionally, the speed itself is set to velo.
     *
     * The units of the inputs are engineering units (EGU) per second (e.g. mm/s
     * if the EGU is mm).
     *
     * If the given configuration is invalid (min > max, velo < min, velo > max)
     * and the motor is configured as a variable speed motor (param lib entry
     * motorCanSetSpeed_ is 1), this function returns an asynError.
     *
     * @param velo                Actual velocity (EGU / s)
     * @param vbas                Minimum allowed velocity (EGU / s)
     * @param velo                Maximum allowed velocity (EGU / s)
     *
     * @return asynStatus
     */
    virtual asynStatus setVeloFields(double velo, double vbas, double vmax);

    /**
     * @brief Populate the ACCL field of the motor record
     *
     * Populates the acceleration field of the motor record with the given
     * value. If accl is not positive, this function does not set the value and
     * returns an asynError.
     *
     * The unit of the input is engineering units (EGU) per second squared (e.g.
     * mm/s^2 if the EGU is mm).
     *
     * @param accl                Actual acceleration (EGU / s^2)
     * @return asynStatus
     */
    virtual asynStatus setAcclField(double accl);

    /**
     * @brief Start the watchdog for the movement, if the watchdog is not
    disabled. See the documentation of checkMovTimeoutWatchdog for more details.
     *
     * @return asynStatus         If one of the parameter library operations
    used to get the values for the timeout calculation failed, return that
    status, otherwise return asynSuccess.
     */
    asynStatus startMovTimeoutWatchdog();

    /**
     * @brief Check if the watchdog timed out
     *
    Manages a timeout mechanism for the movement:
    If the axis is moving and the movement takes too long, create an error
    message and return asynError. The watchdog is started when moving switches
    from "false" to "true" and stopped when moving switches from "true" to
    "false". At the watchdog start, the estimated movement time is calculated as

    t = offsetMovTimeout_ + scaleMovTime_ * [timeContSpeed + 2*timeAccel]

    with

    timeContSpeed = abs(motorTargetPosition - motorPosition) / motorVelBase
    timeAcc = motorVelBase / motorAccel

    The values motorTargetPosition, motorVelBase, motorAccel and
    positionAtMovementStart are taken from the parameter library. Therefore it
    is necessary to populate them before using this function. If they are not
    given, both speed and velocity are assumed to be infinite. This means that
    timeContSpeed and/or timeAcc are set to zero. motorTargetPosition is
    populated automatically when using the doMove function.

    The values offsetMovTimeout_ and scaleMovTimeout_ can be set directly from
    the IOC shell with the functions setScaleMovTimeout and setOffsetMovTimeout,
    if sinqMotor is loaded via the "require" mechanism.
     *
     * @param moving              Should be the "moving" status from `poll` /
    `doPoll`.
     * @return asynStatus         Return asynError, if the watchdog timed out,
    and asynSuccess otherwise.
     */
    asynStatus checkMovTimeoutWatchdog(bool moving);

    /**
     * @brief Enable / disable the watchdog. Also available in the IOC shell
     * (see "extern C" section in sinqController.cpp).
     *
     * If enable is set to false and the watchdog is currently running, this
     * function stops it immediately.
     *
     * @param enabled
     * @return asynStatus
     */
    virtual asynStatus setWatchdogEnabled(bool enable);

    /**
     * @brief Set the offsetMovTimeout. Also available in the IOC shell
     * (see "extern C" section in sinqController.cpp).
     *
    See documentation of `checkMovTimeoutWatchdog` for details.
     *
     * @param offsetMovTimeout    Offset (in seconds)
     * @return asynStatus
     */
    virtual asynStatus setOffsetMovTimeout(time_t offsetMovTimeout) {
        offsetMovTimeout_ = offsetMovTimeout;
        return asynSuccess;
    }

    /**
     * @brief Set the scaleMovTimeout. Also available in the IOC shell
     * (see "extern C" section in sinqController.cpp).
     *
    See documentation of `checkMovTimeoutWatchdog` for details.
     *
     * @param scaleMovTimeout     Scaling factor (in seconds)
     * @return asynStatus
     */
    virtual asynStatus setScaleMovTimeout(time_t scaleMovTimeout) {
        scaleMovTimeout_ = scaleMovTimeout;
        return asynSuccess;
    }

    friend class sinqController;

  protected:
    bool initial_poll_;
    int init_poll_counter_;

    // Helper variables for movementTimeoutWatchdog
    time_t expectedArrivalTime_;
    time_t offsetMovTimeout_;
    double scaleMovTimeout_;
    bool watchdogMovActive_;

  private:
    sinqController *pC_;
};

#endif
